import re
from ..models import Relation, Entity


class RelationAnnotator:
    def display_entity(self, entity: Entity, position: int) -> str:
        return f'<e{str(position)}>{entity.text}</e{str(position)}>'

    def annotate(self, text: str, relation: Relation, offset = 0) -> str:
        annotated_text = text[:]

        e1 = relation.e1
        e1_start = e1.location.start - offset
        e1_end = e1.location.end - offset

        e2 = relation.e2
        e2_start = e2.location.start - offset
        e2_end = e2.location.end - offset

        if e1_end < e2_start:
            return annotated_text[:e1_start] + \
                self.display_entity(e1, 1) + \
                annotated_text[e1_end:e2_start] + \
                self.display_entity(e2, 2) + \
                annotated_text[e2_end:]

        return annotated_text[:e2_start] + \
            self.display_entity(e2, 2) + \
            annotated_text[e2_end:e1_start] + \
            self.display_entity(e1, 1) + \
            annotated_text[e1_end:]

class RelationAnnotatorWithEntityType(RelationAnnotator):
    def display_entity(self, entity: Entity, position: int) -> str:
        return f'<e{str(position)}:{entity.label}>{entity.text}</e{str(position)}:{entity.label}>'

class HtmlRelationAnnotator(RelationAnnotator):
    def display_entity(self, entity: Entity, position: int) -> str:
        key = re.sub(r' ', '-', entity.label)
        return f'<span class="entity lb-{key} e{position}">' + \
            f'<span class="label">{entity.label}</span>' + \
            f'{entity.text}' + \
            '</span>'



from typing import List, Optional
from dataclasses import dataclass

from .annotator import EntityAnnotator
from ..models import Entity
from ..utils import Query
from ..regexes import SlimRegEx


@dataclass
class AttributeSetup:
    before: Optional[str] = None
    after: Optional[str] = None

@dataclass
class AttributeApplications:
    entities: List[str]
    setups: List[AttributeSetup]

    def get_expression(self):
        expressions = []
        for setup in self.setups:
            expression = setup.before if setup.before else ''
            expression += r'(##ENTITY_(?:' + '|'.join(self.entities) + r')_\d+##)'
            if setup.after:
                expression += setup.after

            expressions.append(expression)

        return SlimRegEx(expressions)

@dataclass
class AttributeToApply:
    name: str
    applications: List[AttributeApplications]

class EntityAttributor:
    def __init__(self, attribute_name: str, settings: List[AttributeToApply]) -> None:
        self._attribute_name = attribute_name
        self._settings = settings

    def set_attributes(self, text: str, entities: List[Entity]) -> List[Entity]:
        annotated_text = EntityAnnotator().annotate(text, entities).annotated_text

        mappings = Query(entities).todict(str)
        for setting in self._settings:
            for application in setting.applications:
                for match in application.get_expression().findall(annotated_text):
                    groups = list(match.groups())
                    key = Query(groups) \
                        .find(lambda g: g.startswith('##ENTITY_'))

                    if key:
                        mappings[key].add_attribute(
                            self._attribute_name,
                            setting.name
                        )

        return list(mappings.values())



from typing import List, Optional

from ...models import Relation
from ...entities.viewers.html import __TEMPLATE__
from ..annotator import HtmlRelationAnnotator

class HtmlViewer:
    __TEMPLATE = __TEMPLATE__ \
        .replace(
            '{{body}}',
            """        <table>
{{body}}
        </table>"""
        )

    __DEFAULT_STYLES = """
table { width: 100%; }
span.entity { border: 1px solid black; border-radius: 5px; padding: 5px; margin: 3px; cursor: pointer; }
span.label { font-weight: bold; padding: 3px; color: black; }
span.e1 { background-color: aqua; }
span.e2 { background-color: coral; }
tr.delete { background-color: black !important; }
tr.delete span { background-color: black !important; }
td { line-height: 30px; border: 1px solid black; padding: 5px; }
td.header { font-weight: bold; }
td.label { font-weight: bold; text-align: center; }
"""

    def __init__(self, show_index=True) -> None:
        self._relation_index = 0
        self._show_index = show_index
        self._rows: List[str] = []
        self._annotator = HtmlRelationAnnotator()

    def append_header(self, header: str) -> None:
        colspan = 3 if self._show_index else 2
        self._rows.append(
            '<tr>' + \
                f'<td class="header" colspan="{colspan}">{header}</td>' + \
            '</tr>'
        )

    def append_relation(self, text: str, relation: Relation) -> None:
        index = self._relation_index
        annotation = self._annotator.annotate(text, relation)
        self._rows.append(
            f'<tr id="{index}">' + \
                (f'<td>{index}</td>' if self._show_index else '') + \
                f'<td class="label">{relation.label}</td>' + \
                f'<td>{annotation}</td>' + \
            '</tr>'
        )

        self._relation_index += 1

    def create_view(self, custom_styles: Optional[str] = None) -> str:
        styles = self.__DEFAULT_STYLES.strip()
        if custom_styles:
            styles += f'\n{custom_styles.strip()}'

        body = '\n'.join(self._rows)

        return self.__TEMPLATE \
            .replace('{{styles}}', styles) \
            .replace('{{body}}', body) \
            .strip()



from typing import List, Dict, Optional

from .extractor import EntityExtractor
from ..regexes import RegExLabel, transform_knowledge


def create_entity_extractor(regex_labels: List[RegExLabel], kb: Optional[Dict[str, List[str]]] = None):
    all_regex_labels = []

    if len(regex_labels) > 0:
        all_regex_labels.extend(regex_labels)

    if kb:
        all_regex_labels.extend(
            [transform_knowledge(label, expressions) for label, expressions in kb.items()]
        )

    return EntityExtractor(all_regex_labels)



from abc import ABC, abstractmethod
from typing import List, cast

from ..regexes import RegExLabel
from ..utils import flatten
from ..models import Location, Entity


class AbstractEntityExtractor(ABC):
    @abstractmethod
    def get_entities(self, text: str) -> List[Entity]:
        pass

class EntityExtractor(AbstractEntityExtractor):
    def __init__(self, regex_labels: List[RegExLabel]) -> None:
        self._regex_labels = regex_labels

    def get_entities(self, text: str) -> List[Entity]:
        entities = []
        for regex_label in self._regex_labels:
            for label, match in regex_label.findall(text):
                entities.append(
                    Entity(
                        len(entities) + 1,
                        label,
                        match.group(),
                        Location(*match.span())
                    )
                )

        if len(entities) == 0:
            return []

        ## sort descending
        all_found_entities = sorted(
            entities,
            key=lambda entity: (entity.end, -entity.start),
            reverse=True
        )

        entities = all_found_entities[:1]
        for curr_entity in all_found_entities[1:]:
            prev_entity = entities[-1]

            if curr_entity.is_in(prev_entity):
                continue

            entities.append(curr_entity)

        return entities



from .extractor import AbstractEntityExtractor, EntityExtractor
from .annotator import EntityAnnotator, \
                       HtmlEntityAnnotator, \
                       LabelOnlyEntityAnnotator
from .attributor import EntityAttributor, \
                        AttributeToApply, \
                        AttributeApplications, \
                        AttributeSetup
from .factories import create_entity_extractor



from typing import List, Optional

from ...models import Entity
from ..annotator import HtmlEntityAnnotator


__TEMPLATE__ = """
<html>
    <head>
        <style>
{{styles}}
        </style>
    </head>
    <body>
{{body}}
    </body>
</html>
    """


class HtmlViewer:
    __TEMPLATE = __TEMPLATE__

    __DEFAULT_STYLES = """
p { margin: 5px; line-height: 45px; }
span.entity { border: 1px solid black; border-radius: 5px; padding: 5px; margin: 3px; cursor: pointer; }
span.label { font-weight: bold; padding: 3px; color: black; }
"""

    def __init__(self, annotations: Optional[List[str]] = None) -> None:
        self._rows: List[str] = annotations if annotations is not None else []
        self._annotator = HtmlEntityAnnotator()

    def append(self, text: str, entities: List[Entity]) -> None:
        self._rows.append(
            self._annotator.annotate(text, entities).annotated_text
        )

    def create_view(self, custom_styles: Optional[str] = None, spacer='<hr />') -> str:
        styles = self.__DEFAULT_STYLES.strip()
        if custom_styles:
            styles += f'\n{custom_styles.strip()}'

        format_annoations = (f'<p>{annotation}</p>' for annotation in self._rows)
        body = f'{spacer}\n'.join(format_annoations)

        return self.__TEMPLATE \
            .replace('{{styles}}', styles) \
            .replace('{{body}}', body) \
            .strip()



import setuptools


with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='extr',
    version='0.0.36',
    keywords=['Named Entity Recognition', 'Relation Extraction', 'NER', 'RE', 'NLP'],
    description='Named Entity Recognition (NER) and Relation Extraction (RE) library using Regular Expressions',
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    install_requires=[],
    url='https://github.com/dpasse/extr',
    long_description=long_description,
    long_description_content_type='text/markdown',
)



from .extractor import AbstractRelationExtractor, RelationExtractor
from .annotator import RelationAnnotator, RelationAnnotatorWithEntityType, HtmlRelationAnnotator
from .label_builder import RegExRelationLabelBuilder, RelationLabelBuilderConfig
